unit SourceList;

interface

uses Windows,
     Messages,
     SysUtils,
     Classes,
     Graphics,
     StdCtrls,
     ExtCtrls,
     Controls,
     Forms,
     Splines,
     Printers,
     Math,
     Dialogs;

const CR                   = #13;
      LF                   = #10;
      EOL                  = CR+LF;
      NoWords              = 59;
      ReservedWords        : array[1..NoWords] of string[14] = ('AND','ASM','ARRAY','BEGIN','CASE','CLASS','CONST','CONSTRUCTOR',
                                                                'DESTRUCTOR','DIV','DO','DOWNTO','ELSE','END','EXPORTS','FILE',
                                                                'FOR','FUNCTION','GOTO','IF','IMPLEMENTATION','IN','INHERITED',
                                                                'INTERFACE','LABEL','LIBRARY','MOD','NIL','NOT','OBJECT','OF',
                                                                'OR','OVERRIDE','PACKED','PRIVATE','PROCEDURE','PROGRAM','PROPERTY',
                                                                'PROTECTED','PUBLIC','PUBLISHED',
                                                                'RECORD','REPEAT','SET','SHL','SHR','STATIC','STRING','THEN','TO','TYPE',
                                                                'UNIT','UNTIL','USES','VAR','VIRTUAL','WHILE','WITH','XOR');
      Operators            : set of Char = [':','=','-','+','*','/'];
      ValidCharacters      : Set of Char = ['a'..'z','A'..'Z','0'..'9'];

type TSourceList           = class;
     THeader               = class(TPersistent)
                                private
                                   FOwner : TSourceList;
                                   FFont  : TFont;
                                   FShow  : Boolean;
                                   FText  : String;
                                   FColor : TColor;
                                   FPosition : Integer;
                                   FWidth    : Integer;
                                   procedure FSetFont(Val:TFont);
                                   procedure FSetShow(Val:Boolean);
                                   function  FGetHeight:Integer;
                                   procedure FSetColor(Val:TColor);
                                   procedure FSetText(Val:String);
                                public
                                   constructor Create(Owner:TSourceList;Col:TColor);
                                   destructor  Destroy;
                                   procedure   Clear;
                                   procedure   Draw(PageNr:Integer);
                                published   
                                   property Font   : TFont read FFont write FSetFont;
                                   property Show   : Boolean read FShow write FSetShow;
                                   property Height : integer read FGetHeight;
                                   property Color  : TColor read FColor write FSetColor;
                                   property Text   : String read FText write FSetText;
                             end;
     TSourceList           = class(TCustomPanel)
                                private
                                   FSource         : TStringList;
                                   FHeader         : THeader;
                                   FFooter         : THeader;
                                   FFile           : TextFile;
                                   FFileName       : string;
                                   FLastWord       : string;
                                   FPageNumber     : Integer;
                                   FHeight,FWidth  : Integer;
                                   FTop,FBottom    : Integer;
                                   FMaxFontWidth   : Integer;
                                   FMaxFontHeight  : Integer;
                                   FDrawingCanvas  : TCanvas;
                                   FFont           : TFont;
                                   FVertScrollbar  : TScrollBar;
                                   FComment        : Boolean;
                                   FText           : Boolean;
                                   FReserved       : Boolean;
                                   FOperator       : Boolean;
                                   FInitialized    : boolean;
                                   FOneLineComment : Boolean;
                                   FShowLineNumbers: Boolean;
                                   FPrinting       : Boolean;
                                   FNumberOfLines  : Integer;  // No. of lines on 1 page
                                   FNumberOfPages  : Integer;  // Tot. number of pages needed
                                   function        FGetFileName:string;
                                   procedure       FSetFileName(const value:string);
                                   Procedure       FOnVertScrollBarChange(Sender: TObject);
                                   procedure       FInitialize;
                                   function        FReservedWord(var Input:String):Boolean;
                                   function        FIsOperator(Input:Char):Boolean;
                                   function        FValidCharacter(Input:Char):Boolean;
                                   function        FGetDrawingCanvas:TCanvas;
                                   procedure       FSetShowLineNumbers(Val:Boolean);
                                   procedure       FDrawPage(Nr:Integer);
                                   function        FGetFond:TFont;
                                   procedure       FSetFont(Val:TFont);
                                public
                                   constructor Create(AOwner: TComponent);                             override;
                                   destructor  Destroy;                                                override;
                                   procedure   Clear;                                                  virtual;
                                   property    DrawingCanvas:TCanvas read FGetDrawingCanvas;
                                   procedure   LoadFromFile;
                                   procedure   Paint;override;
                                   procedure   Resize;override;
                                   procedure   Print;
                                published
                                   property Align;
                                   property BevelInner;
                                   property BevelOuter;
                                   property Caption;
                                   property Color;
                                   property Font:TFont read FGetFond write FSetFont;
                                   property OnResize;
                                   property Visible;
                                   property FileName:string read FGetFileName write FSetFilename;
                                   property ShowLineNumbers:Boolean read FShowLineNumbers write FSetShowLineNumbers;
                                   property Header:THeader read FHeader write FHeader;
                                   property Footer:THeader read FFooter write FFooter;
                             end;



procedure Register;


implementation

procedure THeader.FSetFont(Val:TFont);
begin
   FFont.Assign(Val);
   FOwner.Invalidate;
end;{THeader.FSetFont}

procedure THeader.FSetShow(Val:Boolean);
begin
   if Val<>Show then
   begin
      FShow:=Val;
      FOwner.Invalidate;
   end;
end;{THeader.FSetShow}

function THeader.FGetHeight:Integer;
begin
   Result:=Round(1.5*FOwner.DrawingCanvas.TextHeight('ABCDEFGHIJKLMNOPQRSTUVWXYZ12345678990'));
end;{THeader.FGetHeight}

procedure THeader.FSetColor(Val:TColor);
begin
   if Val<>FColor then
   begin
      FColor:=Val;
      FOwner.Invalidate;
   end;
end;{THeader.FSetColor}

procedure THeader.FSetText(Val:String);
begin
   if Val<>Text then
   begin
      FText:=Val;
      FOwner.Invalidate;
   end;
end;{THeader.FSetText}

constructor THeader.Create(Owner:TSourceList;Col:TColor);
begin
   inherited Create;
   FOwner:=Owner;
   FFont:=TFont.Create;
   FFont.name:='ARIAL';
   FFont.Size:=12;
   FFont.Color:=clBlue;
   FColor:=Col;
   FShow:=True;
   Clear;
end;{THeader.Create}

destructor THeader.Destroy;
begin
   FFont.Destroy;
   Inherited Destroy;
end;{THeader.Destroy}

procedure THeader.Clear;
begin
   FText:='';
end;{THeader.Clear}

procedure THeader.Draw(PageNr:Integer);
var H : Integer;
    Str : string;
begin
   if Show then
   begin
      H:=Height;
      FOwner.DrawingCanvas.Font.Assign(Font);
      FOwner.DrawingCanvas.Brush.Color:=Color;
      FOwner.DrawingCanvas.Brush.Style:=bsSolid;
      FOwner.DrawingCanvas.Pen.Color:=Color;
      FOwner.DrawingCanvas.Rectangle(0,FPosition,FWidth,FPosition+H);
      FOwner.DrawingCanvas.TextOut(FOwner.DrawingCanvas.TextWidth('X'),FPosition+(H-FOwner.DrawingCanvas.TextHeight(FText)) div 2,FText);
      if PageNr>0 then
      begin
         FOwner.DrawingCanvas.Font.Size:=round(0.5*FOwner.DrawingCanvas.Font.Size);
         if FOwner.DrawingCanvas.Font.Size<7 then FOwner.DrawingCanvas.Font.Size:=7;
         FOwner.DrawingCanvas.Font.Color:=clBlack;
         Str:='Page '+IntToStr(PageNr)+'/'+IntToStr(FOwner.FVertScrollbar.Max);;
         FOwner.DrawingCanvas.TextOut(FWidth-round(2*FOwner.DrawingCanvas.TextWidth(Str)),FPosition+(H-FOwner.DrawingCanvas.TextHeight(Str)) div 2,Str);
      end;

      FOwner.DrawingCanvas.Brush.Style:=bsClear;
      FOwner.DrawingCanvas.Pen.Color:=clBlack;
      FOwner.DrawingCanvas.Pen.Width:=2;
      FOwner.DrawingCanvas.MoveTo(0,FPosition-2);
      FOwner.DrawingCanvas.LineTo(FWidth,FPosition-2);
      FOwner.DrawingCanvas.MoveTo(0,FPosition+H+2);
      FOwner.DrawingCanvas.LineTo(FWidth,FPosition+H+2);
   end;
end;{THeader.Draw}

{--------------------------------------------------------------------------------------------------}
{                                           TSourceList                                            }
{--------------------------------------------------------------------------------------------------}
function TSourceList.FGetFileName:string;
begin
   Result:=FFileName;
end;{TSourceList.FGetFileName}

procedure TSourceList.FSetFileName(const value:string);
begin
   FFileName:=value;
end;{TSourceList.FSetFileName}

Procedure TSourceList.FOnVertScrollBarChange(Sender: TObject);
begin
   Invalidate;
end;{TSourceList.FOnVertScrollBarChange}

procedure TSourceList.FInitialize;
var I,W,H : Integer;
begin
   if FPrinting then
   begin
      FHeight:=Printer.PageHeight;
      FWidth:=Printer.PageWidth;
   end else
   begin
      FHeight:=ClientHeight;
      FWidth:=ClientWidth;
   end;
   if Header<>nil then
   begin
      Header.FPosition:=0;
      Header.FWidth:=FWidth;
      if Header.Show then FTop:=Header.Height
                     else FTop:=0;
   end else FTop:=0;
   if Footer<>nil then
   begin
      Footer.FPosition:=FHeight-Footer.Height;
      Footer.FWidth:=FWidth;
      if Footer.Show then FBottom:=FHeight-Footer.Height
                     else FBottom:=FHeight;
   end else FBottom:=FHeight;
   Drawingcanvas.Font.Assign(Font);
   FMaxFontWidth:=0;
   FMaxFontHeight:=0;
   For I:=1 to 255 do
   begin
       W:=DrawingCanvas.TextWidth(Chr(I));
       H:=DrawingCanvas.TextHeight(Chr(I));
       if W>FMaxFontWidth then FMaxFontWidth:=W;
       if H>FMaxFontHeight then FMaxFontHeight:=H;
   end;
   if FMaxFontHeight=0 then FMaxFontHeight:=4;
   FNumberOfLines:=((FBottom-FTop) div FMaxFontHeight)-1;
   if FNumberOfLines<1 then FNumberOfLines:=1;
   FNumberOfPages:=1+(FSource.Count div FNumberOfLines);
   FVertScrollBar.LargeChange:=1;
   FVertScrollBar.Max:=FNumberOfPages;
   FInitialized:=True;
end;{TSourceList.FInitialize}

function TSourceList.FGetDrawingCanvas:TCanvas;
begin
   if FPrinting then Result:=Printer.Canvas
                else Result:=Canvas;
end;{TSourceList.FGetDrawingCanvas}

function TSourceList.FReservedWord(var Input:String):Boolean;
var I   : Integer;
    Tmp : String;
begin
   Result:=False;
   Tmp:=Uppercase(Input);
   for I:=1 to NoWords do
   begin
      if ReservedWords[I]=Tmp then
      begin
         Input:=LowerCase(Input);
         Result:=True;
         Break;
      end;
   end;
end;{TSourceList.FReservedWord}

function TSourceList.FIsOperator(Input:Char):Boolean;
begin
   Result:=Input in Operators;
end;{TSourceList.FIsOperator}

function TSourceList.FValidCharacter(Input:Char):Boolean;
begin
   Result:=Input in ValidCharacters;
end;{TSourceList.FValidCharacter}

procedure TSourceList.FSetShowLineNumbers(Val:Boolean);
begin
   if Val<>ShowLineNumbers then
   begin
      FShowLineNumbers:=Val;
      Invalidate;
   end;
end;{TSourceList.FSetShowLineNumbers}

procedure TSourceList.FDrawPage(Nr:Integer);
var Index   : Integer;
    Str : String;
    XPos  : Integer;
    YPos  : Integer;

    procedure WriteToCanvas(Str:String);
    begin
       if FText then
       begin
          DrawingCanvas.Font.Color:=clBlue;//$00FF8000;
          DrawingCanvas.Font.Style:=[];
       end else if FComment then
       begin
          DrawingCanvas.Font.Color:=cldkGray;
          DrawingCanvas.Font.Style:=[fsItalic];
       end else if FOneLineComment then
       begin
          DrawingCanvas.Font.Color:=clRed;
          DrawingCanvas.Font.Style:=[fsItalic];
       end else if (FReserved) or (FReservedWord(Str)) then
       begin
          DrawingCanvas.Font.Color:=clBlack;
          DrawingCanvas.Font.Style:=[fsBold];
       end else if FOperator then
       begin
          DrawingCanvas.Font.Color:=$008000FF;//clGreen;
          DrawingCanvas.Font.Style:=[];
       end else
       begin
          DrawingCanvas.Font.Color:=clBlack;
          DrawingCanvas.Font.Style:=[];
       end;
       DrawingCanvas.TextOut(Xpos*FMaxFontWidth,FTop-FMaxFontHeight div 2+YPos*FMaxFontHeight,Str);
       inc(XPos,Length(Str));
    end;

    procedure ProcessLine(Source:string);
    var I,J  : Integer;
        Word : string;
        Ch   : Char;

        function FNextCharacter:Char;
        var Tmp : String;
            I   : Integer;
        begin
           Result:=#0;
           if Source<>'' then Result:=Source[1] else
           I:=Index+1;
           while I<=FSource.Count do
           begin
              Tmp:=FSource.Strings[I-1];
              if Tmp<>'' then
              begin
                 Result:=Tmp[1];
                 Break;
              end else inc(I);
           end;
        end;

    begin
       XPos:=0;
       if ShowLineNumbers then
       begin
          Word:=IntToStr(Index);
          While length(Word)<6 do Insert(#32,Word,1);
          Word:=Word+':';
          Drawingcanvas.Font.Color:=clMaroon;
          Drawingcanvas.Font.Style:=[];

          DrawingCanvas.TextOut(Xpos*FMaxFontWidth,FTop-FMaxFontHeight div 2+YPos*FMaxFontHeight,Word);
          Drawingcanvas.Font.Color:=clBlack;
          inc(XPos,8);
       end;
       Word:='';
       FOneLineComment:=False;
       While Length(Source)>0 do
       begin
          Ch:=Source[1];
          Delete(Source,1,1);
          if FText then // Current mode is textmode
          begin
             Word:=Word+Ch;
             if Ch='''' then
             begin
                WriteToCanvas(Word);
                FText:=False;
                Word:='';
             end;
          end else if (FComment) or (FOneLineComment) then // Current mode is commentmode
          begin
             Word:=Word+Ch;
             if FComment then
             begin
                if Ch='}' then
                begin
                   WriteToCanvas(Word);
                   FComment:=False;
                   Word:='';
                end else if Source='' then
                begin
                   WriteToCanvas(Word);
                end;
             end else if FOneLineComment then
             begin
                if Source='' then WriteToCanvas(Word);
             end;
          end else if FIsOperator(Ch) then // Operator encountered, write word;
          begin
             WriteToCanvas(Word);
             Word:='';
             if (CH='/') then
             begin
                if FNextCharacter='/' then
                begin
                   FOneLineComment:=True;
                   Word:=Ch;
                end else
                begin
                   FOperator:=True;
                   WriteToCanvas(Ch);
                   FOperator:=False;
                   FOneLineComment:=True;
                end;
             end else
             begin
                FOperator:=True;
                WriteToCanvas(Ch);
                FOperator:=False;
             end;
          end else if (Source='') or (not FValidCharacter(CH)) then
          begin
             if FValidCharacter(CH) then
             begin
                Word:=Word+Ch;
                WriteToCanvas(Word);
                Word:='';
             end else if CH=#32 then
             begin
                WriteToCanvas(Word);
                Inc(XPos);
                Word:='';
             end else if CH='{' then
             begin
                WriteToCanvas(Word);
                FComment:=True;
                Word:=Ch;
             end else if CH='''' then
             begin
                WriteToCanvas(Word);
                FText:=True;
                Word:=Ch;
             end else
             begin
               WriteToCanvas(Word);
               Word:='';
               WriteToCanvas(Ch);
             end;
          end else
          begin
             Word:=Word+Ch;
          end;
       end;
    end;

begin
   if Header<>nil then Header.draw(Nr);
   if Footer<>nil then
   begin
      Footer.FText:='Printed '+DateTimeToStr(Now);
      Footer.Draw(0);
   end;
   Index:=1+(Nr-1)*FNumberOfLines;
   Drawingcanvas.Font.Assign(Font);
   YPos:=1;
   XPos:=0;
   While (YPos<=FNumberOfLines) and (Index<=FSource.Count) do
   begin
      Str:=FSource.Strings[Index-1];
      ProcessLine(Str);
      inc(YPos);
      Inc(Index);
   end;
end;{TSourceList.FDrawPage}

function TSourceList.FGetFond:TFont;
begin
   Result:=FFont;
end;{TSourceList.FGetFond}

procedure TSourceList.FSetFont(Val:TFont);
begin
   FFont.Assign(Val);
   FInitialize;
   Invalidate;
end;{TSourceList.FSetFont}

constructor TSourceList.Create(AOwner: TComponent);
begin
   if (csReading in ComponentState) or (csDestroying in ComponentState) then exit;
   inherited Create(AOwner);
   Color:=$00F5F5F5;
   Filename:='';
   FFont:=TFont.Create;
   FFont.Name:='Courier new';
   FFont.Size:=7;
   FSource:=TStringList.Create;
   FInitialized:=False;
   FHeader:=THeader.Create(Self,clSilver);
   FFooter:=THeader.Create(Self,$0080FFFF);
   FVertScrollBar:=TScrollBar.Create(Self);
   FVertScrollBar.Kind:=sbVertical;
   FVertScrollBar.Parent:=Self;
   FVertScrollBar.Visible:=True;
   FVertScrollBar.Min:=1;
   FVertScrollBar.Max:=1;
   FVertScrollBar.LargeChange:=1;
   FVertScrollBar.OnChange:=FOnVertScrollBarChange;
   FVertScrollBar.Width:=16;
   FVertScrollBar.Align:=alRight;
   FPrinting:=false;
   Clear;
end;{TSourceList.Create}

destructor TSourceList.Destroy;
begin
   Clear;
   FSource.Destroy;
   FHeader.Destroy;
   FFooter.Destroy;
   FFont.Destroy;
   inherited destroy;
end;{TSourceList.Destroy}

procedure TSourceList.Clear;
begin
   FLastWord:='';
   FSource.Clear;
end;{TSourceList.Clear}

procedure TSourceList.LoadFromFile;
begin
   if FileExists(FileName) then
   begin
      Clear;
      Header.Text:=Filename;
      FSource.LoadFromFile(FileName);
      FInitialized:=False;
      invalidate;
   end else MessageDlg('File not found.',mtError,[mbOk],0);
end;{TSourceList.LoadFromFile}

procedure TSourceList.Paint;
begin
   inherited Paint;
   if not FInitialized then FInitialize;
   FDrawPage(FVertScrollBar.Position);
end;{TSourceList.Paint}

procedure TSourceList.Resize;
begin
   Finitialized:=False;
   inherited Resize;
end;{TSourceList.Resize}

procedure TSourceList.Print;
var I,Min,Max   : integer;
    Pag         : Integer;
    PrintDialog : TPrintDialog;
begin
   Pag:=FVertScrollbar.Position;
   PrintDialog:=TPrintDialog.Create(Self);
   PrintDialog.Options:=[poPageNums];
   FPrinting:=True;
   Finitialize;
   PrintDialog.MinPage:=1;
   PrintDialog.MaxPage:=FVertScrollbar.Max;
   if PrintDialog.Execute then
   begin
      if PrintDialog.PrintRange=prPageNums then
      begin
         Min:=PrintDialog.FromPage;
         Max:=PrintDialog.ToPage;
      end else
      begin
         Min:=1;
         Max:=FVertScrollbar.Max;
      end;
      For I:=Min to Max do
      begin
         Printer.BeginDoc;
         FDrawPage(I);
         Printer.EndDoc;
      end;
   end;
   FPrinting:=False;
   FInitialize;
   FVertScrollbar.Position:=Pag;
   PrintDialog.Destroy;
end;{TSourceList.Print}

procedure Register;
begin
  RegisterComponents('Martijn', [TSourceList]);
end;{Register}

end.

